/**********************************************************************/
/*  Filename       : DOSIR2PC.C                                       */
/**------------------------------------------------------------------**/
/*  Description    : receive code IR information from COMport         */
/*                   and propagate (map) it to the keyboard buffer    */
/**------------------------------------------------------------------**/
/*  Author         : Jui                                              */
/*  developed at   : 10.03.1999                                       */
/*  last Update    : 12.03.2001                                       */
/*  compiled with borland c++ Rev. 3.1                                */
/**********************************************************************/

//specify the max. keymap buffersize here
#define    MAXCODES 40          

#define    VERSION  " V.1.01"
#define    DATFILE  "ir2pc.dat"
#define    FILENAME "dosir2pc"

#include   <stdlib.h>
#include   <stdio.h>
#include   <conio.h>
#include   <ctype.h>
#include   <string.h>
#include   <bios.h>   
#include   <dos.h>     

// Reduce heaplength and stacklength to make a smaller program in memory
extern unsigned _heaplen = 3256;
extern unsigned _stklen  = 2024;

const   unsigned    BIOSdseg = 0x0040;  // BIOS data segment
const   unsigned    IMR = 0x21;         // Interrupt Mask Register
const   unsigned    safety_space = 5024;// Bytes of spare space

// These are offsets from the UART base port
const   unsigned    UART_THR = 0x00;    // Transmit Hold Register and
const   unsigned    UART_RBR = 0x00;    // Receiver Buffer Register
const   unsigned    UART_IER = 0x01;    // Interrupt Enable Register
const   unsigned    UART_IIR = 0x02;
const   unsigned    UART_LCR = 0x03;    // Line Control Register?
const   unsigned    UART_MCR = 0x04;    // Modem Control Register?
const   unsigned    UART_LSR = 0x05;    // Line Status Register
const   unsigned    UART_MSR = 0x06;

//-----------------------------------------------------------------
// Global variables needed both during initialization and interrupt
// processing.
//-----------------------------------------------------------------

int base_port;  // Base I/O port address
unsigned char   ir_code[MAXCODES][8];
unsigned char   inbuf[6];

/*-------------------------------------------------------------------*/
void interrupt serial_input_handler(void) 
/*-------------------------------------------------------------------*/
{
   static  unsigned char   in_char;
   static  unsigned char   scancode;
   static int found;
   static int t;


   // Get the character from the COM port RBR register
   in_char = inportb(base_port + UART_RBR);
   inbuf[0]=inbuf[1];
   inbuf[1]=inbuf[2];
   inbuf[2]=inbuf[3];
   inbuf[3]=inbuf[4];
   inbuf[4]=inbuf[5];
   inbuf[5]= in_char;

   t=0;
   found=0;
   while ((!found)&&(ir_code[t][0]!='$')) {
       if((inbuf[0] == ir_code[t][2]) &&
          (inbuf[1] == ir_code[t][3]) &&
          (inbuf[2] == ir_code[t][4]) &&
          (inbuf[3] == ir_code[t][5]) &&
          (inbuf[4] == ir_code[t][6]) &&
          (inbuf[5] == ir_code[t][7]) ) found =1; // Got one known code
       else t=t+1;
   }

   if (found) {                                  
      in_char = ir_code[t][0];
      scancode = ir_code[t][1];
      _AH = 0x05;
      _CL = in_char;
      _CH = scancode;
      geninterrupt(0x16);                         // put key to kbuffer
   }
    // Tell the 8259A End Of Interrupt
    outportb(0x20,0x20);
}

/*-------------------------------------------------------------------*/
void  Usage(char *s)
/*-------------------------------------------------------------------*/
{
   fprintf(stderr,"Usage: %s [port] [l]\n",s);
   fprintf(stderr," port: COM port [1-4] (1 = default)\n");
   fprintf(stderr," l   : learning mode \n\n");
   fprintf(stderr,"Plugg in IR2PC modul to specified COM port\n");
   fprintf(stderr,"otherwise this program is useless!\n");
   fprintf(stderr,"Max. number of mappings = %d\n",MAXCODES);
   exit(-1);
}

/*-------------------------------------------------------------------*/
int readkey()
/*-------------------------------------------------------------------*/
{
 register unsigned char i;

      _AH = 0x00;
      geninterrupt(0x16);
      i=_AL;
      return( _AH*256 +i);
}

/*-------------------------------------------------------------------*/
unsigned char Serial_read(void) 
/*-------------------------------------------------------------------*/
{
          while(!((unsigned char) inportb(base_port + UART_LSR)&1)) {}
          return (char)inportb(base_port + UART_RBR);
}

/*-------------------------------------------------------------------*/
void flush_serial(void) 
/*-------------------------------------------------------------------*/
{
      while( (unsigned char)inportb(base_port + UART_LSR)&1) {
        inportb(base_port + UART_RBR);
      }
}

/*-------------------------------------------------------------------*/
void write_IRtable(unsigned char tab[MAXCODES][8])
/*-------------------------------------------------------------------*/
{
  int t = 0;
  int i;
  FILE *fp;

  fp = fopen(DATFILE,"w");
  if (fp == NULL) printf("Error opening file\n");
  while ( tab[t][0] != (unsigned char)255) {
      for (i=0; i<8; i++){
        fprintf( fp,"%02x ",tab[t][i]);
      }   
  fprintf( fp,"\n");
  t++;
  }
  fprintf(fp,"$");
  fclose(fp);
}

/*-------------------------------------------------------------------*/
void read_IRtab(unsigned char tab[MAXCODES][8])
/*-------------------------------------------------------------------*/
{
int t = 0;
int i,l=0;
int value;
FILE *fp;
unsigned char *c;
char input[50];
char * end;

  fp = fopen(DATFILE,"rt");
  if (fp == NULL){
     printf("error opening file %s\n", DATFILE);
     printf("start %s in learningmode first\n",FILENAME);
     printf("Type: %s l\n", FILENAME);

     exit(-1);
  }
  fgets(input,50,fp);

  c=strtok(input," ");
  while (input[0]!='$'){
     l=0;
     while(l<8){
        value=(int)(strtol(c,&end,16));
        tab[t][l] = (char)value;
        c=strtok(NULL," ");
        l++;
     }
     fgets(input,30,fp);
     c=strtok(input," ");
     t++;
  }//while

  fclose(fp);
}

/*-------------------------------------------------------------------*/
void    main(unsigned argc, char *argv[]) 
/*-------------------------------------------------------------------*/
{
   int readkey(void);
   
   int noir = 0;
   int port = 1;       // COM port to read (1-4)
   int temp;           // Temporary holding register
   int int_to_grab;    // Which interrupt vector to grab
   int int_to_enable;  // Which hardware interrupt to enable
   int t;
   int codecount=0;
   char c,s;           
   int  mode = 0;      // 0 default mode
                       // 1 learn mode
   int  stop = 0;
   int  retval;
   clrscr();
   printf("Remote to keyboardmap %s (c) 1999 Jui\n\n",VERSION);
    // Parse the command line
    switch (argc) {
      case 1:   // No arguments
        break;
      case 2:   // COM port number (1-4) as argument
        if (*argv[1]=='l') {
            mode = 1;
            break;
        }
       
      case 3:
        port = atoi(argv[1]);
        if ( (port < 1) || (port > 4) )
             Usage(argv[0]);
        if ((argc==3) && (*argv[2]=='l')) {
            mode = 1;
        }
        break;
      default:
        Usage(argv[0]); // show help
    }
    switch (port) {
      case 1:
      case 3:
        int_to_grab = 0x0c;
        int_to_enable = 4;
        break;
      case 2:
      case 4:
        int_to_grab = 0x0b;
        int_to_enable = 3;
        break;
      default:
        fprintf(stderr,"Internal error 1 - port is %d\n",port);
        exit(-1);
    }

    base_port = *(int far *)MK_FP(BIOSdseg,(port-1)*2);
    if (base_port == 0) {
        fprintf(stderr,"COM%d has no BIOS base port\n",port);
        fprintf(stderr,"    (%s not installed)\n",FILENAME);
        exit(-1);
    }

    // Set the parameters on the COM port to:
    //  0xE0:   9600 Baud
    //  0x03:   8 Data bits (0x02 = 7 data bits)
    //  0x00:   1 Stop bit  (0x04 = 2 stop bits)
    //  0x00:   No parity   (0x08 = odd, 0x18 = even)
    (void)bioscom(0, 0xE0 | 0x03 | 0x00 | 0x00 , port-1);

    /* install the new interrupt handler for serial input */
    setvect(int_to_grab, serial_input_handler);

    // Disable interrupts while we are messing with the UART
    disable();

    //-----------------------------------------------------------------
    // Program the UART for the serial port we are using
    //-----------------------------------------------------------------

    // Turn off the Divisor Access Latch Bit to allow access to RBR, etc.
    temp = inportb(base_port + UART_LCR) & 0x7F;  // Strip upper bit
    outportb(base_port + UART_LCR, temp);

    // Assert DTR, RTS, and OUT2 on the COM port
    outportb(base_port + UART_MCR, 0x08);

    // Enable interrupts for data ready on the serial line (8250)
    outportb(base_port + UART_IER, 0x01);

    // Read the Line Status Register to clear any reported errors
    temp = inportb(base_port + UART_LSR);

    // Read the Receiver Buffer Register to clear any incoming character
    temp = inportb(base_port + UART_RBR);

   printf ("Check communication at COM%d.\nPlease wait or press Enter for no IR2PC modul.....\n",port);
   temp = 0;
   while ( (unsigned char)temp != 'X'  ) {
        outportb(base_port + UART_MCR, 0x08);
        delay(1000);
        outportb(base_port + UART_MCR, 0x0B);
        delay(1000);
        if (kbhit()) {
            noir=1;
            break;
        }

        if (((unsigned char) inportb(base_port + UART_LSR)&1)) {
            temp = inportb(base_port + UART_RBR);
            if (kbhit()) {
                noir=1;
                break;
            }
        }
   }
   if (noir){
       printf("No IR2PC modul available\n");
       Usage(argv[0]);
       exit(-1);

   }
   // Initialisation - write 'IR' to Com port
   printf ("Communication ok\n");
   outportb(base_port + UART_THR, (int)'I');
   delay (100);
   outportb(base_port + UART_THR, (int)'R');

   if (mode == 1){
      printf("##### Learning mode ######\n");
      printf("01. Press keyboard key to simulate by remote control, <ALT-q> for quit\n");
      retval=readkey();            // get key from keyboard
      if(retval==0x1000) stop=1;   // thats alt-q
      while (!stop) {
          c = (retval & 0xff);
          s = (retval & 0xff00)/256;
          printf("    Press remote button\n");
          flush_serial();
          ir_code[codecount][0] = c;
          ir_code[codecount][1] = s;
          printf ("Fetch code for <%02x %02x>: ",c,s);
          for (t=2; t<8;t++){
              ir_code[codecount][t] = Serial_read();
              printf ("%02x ",ir_code[codecount][t] );
          }
          printf ("\n\n");
          codecount = codecount+1;
          if(codecount == MAXCODES) break; // end of buffer 
          printf("%02d. Press keyboard key to simulate by remote control, <ALT-q> for quit\n",codecount+1);
          retval=readkey();
          if(retval==0x1000) stop=1;       // alt-q     
      }
      ir_code[codecount][0] = (unsigned char)255;
      write_IRtable(ir_code);
      printf("\n%d Keymaps fetched and stored in file: %s\n",codecount,DATFILE); 

    }
    else {
     // read code map
     read_IRtab(ir_code);
     
     //-----------------------------------------------------------------
     // Enable the IRQ for the serial line on the 8259A interrupt control
     // Do this by clearing the bit in the ISR corresponding to the
     // interrupt that is to be enabled.
     //-----------------------------------------------------------------

     temp = inportb(IMR);            // Read the value of the Interrupt Mask Reg
     temp &= ~(1 << int_to_enable);  // Zero bit for one to enable
     outportb(IMR,temp);             // Send the value back

     // Re-enable interrupts, since all is set up
     enable();
     printf ("TSR successful installed \n");
     // Terminate and Stay Resident
     // now you can press your remote buttons, the appropriate key will be displayed
     // on the screen, or transported to other programs.  
     keep(0,(_SS+( (_SP+safety_space)/16)-_psp));
    }
}
